{: Background worker thread and work classes that they manipulate.}
unit WorkUnit;

interface

uses Windows, Messages, SysUtils, Classes, Forms ;

const
  {: Windows user message used to notify a form that work has been completed.
     Make sure this doesn't clash with any other user messages.}
  WM_WORKCOMPLETE = WM_USER + 1001 ;

type
  {: The base class for work that needs to be performed in the background.
     This work is performed by a worker thread of type <see class=TWorkerThread>.}
  TWork = class
  private
    FExceptionClass: ExceptClass;
    FExceptionMessage: string;
    FIdentification: string;
  protected
    procedure ExceptionRaised( const E : Exception ) ;
    procedure SetIdentification( const AIdentification : string ) ;

    {: Override this to perform the work task.}
    procedure Perform ; virtual ; abstract ;
  public
    constructor Create( const AIdentification : string ) ;
    property  ExceptionMessage : string read FExceptionMessage ;
    property  ExceptionClass : ExceptClass read FExceptionClass ;
    property  Identification : string read FIdentification ;
  end ;

  {: A thread-safe list of <see class=TWork> with semaphores that the
     <see class=TWorkerThread> class uses to control removal and
     addition of items to the list.}
  TWorkList = class
  private
    FList : TList ;
    FCritSect : TRTLCriticalSection ;
    FWorkAvailable: THandle;
    FSpaceAvailable: THandle;
    function  GetCount: integer;
    function  GetItem(const AIndex: integer): TWork;
    function  CreateSemaphore( const ACount : integer ) : THandle ;
    function  MultiWaitFor( const ASyncObject       : THandle ;
                            const ATerminationEvent : THandle ) : boolean ;
    procedure SpaceAdded ;
    procedure WorkAdded ;
  public
    constructor Create( const AMaxItems : integer ) ;
    destructor Destroy ; override ;
    procedure Lock ;
    procedure Unlock ;
    procedure Add( const AWork : TWork ) ;
    procedure Delete( const AIndex : integer ) ;
    procedure Remove( const AWork : TWork ) ;
    function  WaitForWork( const ATerminationEvent : THandle ) : TWork ;
    function  WaitAndAdd( const ATerminationEvent : THandle ;
                          const AWork             : TWork ) : boolean ;
    property  Count : integer read GetCount ;
    property  Items[ const AIndex : integer ] : TWork read GetItem ; default ;
    property  WorkAvailable : THandle read FWorkAvailable ;
    property  SpaceAvailable : THandle read FSpaceAvailable ;
  end ;

  TWorkerThread = class ;  // forward declaration

  TWorkStatus = ( wsStart, wsComplete ) ;

  {: Called when a worker thread starts to do any work. Synchronised with the
     main thread.}
  TWorkEvent = procedure( const Sender      : TWorkerThread ;
                          const AWork       : TWork ;
                          const AWorkStatus : TWorkStatus ) of object ;

  {: Thread that does work of type <see class=TWork.>}
  TWorkerThread = class( TThread )
  private
    FNotificationForm: TForm;
    FTerminationEvent: THandle;
    FCompletionList: TWorkList;
    FWorkList: TWorkList;
    FOnWork: TWorkEvent;
    FWork: TWork;
    FWorkStatus : TWorkStatus ;
    procedure DoingWork( const AWork       : TWork ;
                         const AWorkStatus : TWorkStatus ) ;
  protected
    procedure CallOnWork ;
    function  GetWork : TWork ;
    procedure WorkComplete( const AWork : TWork ) ;
    procedure Execute ; override ;
  public
    constructor Create( const ACreateSuspended  : boolean ;
                        const AWorkList         : TWorkList ;
                        const ACompletionList   : TWorkList ;
                        const ANotificationForm : TForm ;
                        const ATerminationEvent : THandle ) ;
    property  WorkList : TWorkList read FWorkList ;
    property  CompletionList : TWorkList read FCompletionList ;
    property  NotificationForm : TForm read FNotificationForm ;
    property  TerminationEvent : THandle read FTerminationEvent ;
    property  OnWork : TWorkEvent read FOnWork write FOnWork ;
  end ;

implementation

{ TWorkerThread }

{: Notes the work list, completion list, notification form and
   termination event, then calls the inherited constructor to
   start the thread.}
constructor TWorkerThread.Create(const ACreateSuspended: boolean;
  const AWorkList, ACompletionList: TWorkList;
  const ANotificationForm: TForm; const ATerminationEvent: THandle);
begin
  // check parameters
  Assert( Assigned( AWorkList ) and Assigned( ACompletionList ),
          'Work list and/or completion list not assigned' ) ;
  Assert( Assigned( ANotificationForm ), 'notification form not assigned' ) ;
  Assert( ATerminationEvent <> 0, 'termination event is zero (not assigned)' ) ;

  // note parameters
  FWorkList := AWorkList ;
  FCompletionList := ACompletionList ;
  FNotificationForm := ANotificationForm ;
  FTerminationEvent := ATerminationEvent ;

  // call inherited at the end of the end of the constructor because
  // this will cause the thread to start running - we need to be sure that
  // initialisation is complete before doing this
  inherited Create( ACreateSuspended ) ;
end;

{: Synchronised call for the OnWork event.}
procedure TWorkerThread.CallOnWork;
begin
  FOnWork( Self, FWork, FWorkStatus ) ;
end;

{ TWork }

constructor TWork.Create(const AIdentification: string);
begin
  inherited Create ;
  FIdentification := AIdentification ;
end;

{: Notes the exception string and class.}
procedure TWork.ExceptionRaised(const E: Exception);
begin
  FExceptionMessage := E.Message ;
  FExceptionClass := ExceptClass( E.ClassType ) ;
end;

{: Sets the work's identification string for display, debugging, etc.
   This protected method is for use by subclasses to change the
   identification string as the work proceeds if necessary,
   since the identification string is passed to the constructor.}
procedure TWork.SetIdentification(const AIdentification: string);
begin
  FIdentification := AIdentification ;
end;

{ TWorkList }

{: Creates the internal list used to hold work, initialises
   the critical section used to control access to the list and
   two semaphores for threads to use to synchronise use of the
   list.}
constructor TWorkList.Create(const AMaxItems: integer);
begin
  inherited Create ;

  // create the internal list
  FList := TList.Create ;

  // initialise the critical section
  InitializeCriticalSection( FCritSect ) ;

  // create the synchronisation semaphores
  FWorkAvailable := CreateSemaphore( 0 ) ;
  FSpaceAvailable := CreateSemaphore( AMaxItems ) ;
end;

{: Adds an item of work to the list and signals the work available semaphore.
   Worker threads should do a wait on the <see property=SpaceAvailable> semaphore
   before adding work to the list.}
procedure TWorkList.Add(const AWork: TWork);
begin
  Lock ;
  try
    FList.Add( AWork ) ;
    WorkAdded ;
  finally
    Unlock ;
  end ;
end;

{: Creates an unnamed semaphore with the given initial count.
   Raises an exception if the creation failed - unlikely but...}
function TWorkList.CreateSemaphore(const ACount: integer): THandle;
begin
  Result := Windows.CreateSemaphore( NIL, ACount, MaxInt, NIL ) ;
  if Result = 0 then
    RaiseLastWin32Error ;
end;

{: Frees the internal list.}
destructor TWorkList.Destroy;
begin
  // we don't own work, so simply free the internal list of work
  FList.Free ;
  inherited;
end;

{: Returns the count of work in the list.}
function TWorkList.GetCount: integer;
begin
  Lock ;
  try
    Result := FList.Count ;
  finally
    Unlock ;
  end ;
end;

{: Returns the item of work at the given index.}
function TWorkList.GetItem(const AIndex: integer): TWork;
begin
  Lock ;
  try
    Result := TWork( FList[ AIndex ] ) ;
  finally
    Unlock ;
  end ;
end;

{: Locks the internal list.}
procedure TWorkList.Lock;
begin
  EnterCriticalSection( FCritSect ) ;
end;

{: Waits for either the sync object or the termination event. Returns
   true if the wait was successful on the sync object and false if
   the wait was completed because of the termination event or an error.}
function TWorkList.MultiWaitFor(const ASyncObject,
  ATerminationEvent: THandle): boolean;
var AHandles : array[ 0..1 ] of THandle ;
begin
  AHandles[ 0 ] := ASyncObject ;
  AHandles[ 1 ] := ATerminationEvent ;
  Result := WaitForMultipleObjects( 2, @AHandles, false, INFINITE ) = WAIT_OBJECT_0 ;
end;

{: Removes an item from the internal list.}
procedure TWorkList.Remove(const AWork: TWork);
begin
  Lock ;
  try
    FList.Remove( AWork ) ;
    SpaceAdded ;
  finally
    Unlock ;
  end ;
end;

{: Unlocks the internal list.}
procedure TWorkList.Unlock;
begin
  LeaveCriticalSection( FCritSect ) ;
end;

{: Runs the thread loop which takes work from the work list,
   executes it then adds it to the completion list and posts a
   completion message to the notification form.}
procedure TWorkerThread.DoingWork(const AWork: TWork;
  const AWorkStatus: TWorkStatus);
begin
  if not Assigned( FOnWork ) then exit ;
  FWork := AWork ;
  FWorkStatus := AWorkStatus ;
  Synchronize( CallOnWork ) ;
end;

procedure TWorkerThread.Execute;
var AWork : TWork ;
begin
  try
    while not Terminated do begin
      // try to get some work - NIL means we've been asked to terminate
      AWork := GetWork ;
      if not Assigned( AWork ) then break ;

      // perform the work, noting any exception raised
      DoingWork( AWork, wsStart ) ;
      try
        try
          AWork.Perform ;
        except
          on E: Exception do
            AWork.ExceptionRaised( E ) ;
        end ;
      finally
        DoingWork( AWork, wsComplete ) ;
      end ;

      // put the work on the completion list
      WorkComplete( AWork ) ;
    end ;
  except
    // squash any exceptions that get this far
  end ;
end;

{: Waits until work is available using the semaphore. If the termination
   event is signalled before this it returns NIL.
   <b>This function should only be called by a background thread.</b>}
function TWorkerThread.GetWork: TWork;
begin
  Result := FWorkList.WaitForWork( FTerminationEvent ) ;
end;

{: Attempts to adds work to the completion list and posts a notification
   message to the notification form if successful.}
procedure TWorkerThread.WorkComplete(const AWork: TWork);
begin
  if FCompletionList.WaitAndAdd( FTerminationEvent, AWork ) then
    PostMessage( FNotificationForm.Handle, WM_WORKCOMPLETE, 0, integer( Self ) ) ;
end;

{: Waits for space on the list using the space available semaphore and termination event,
   then if successful adds AWork to the list. Returns true on success -
   false implies that thread should terminate.}
function TWorkList.WaitAndAdd(const ATerminationEvent : THandle ;
  const AWork: TWork): boolean;
begin
  Result := MultiWaitFor( FSpaceAvailable, ATerminationEvent ) ;
  if not Result then exit ;
  Lock ;
  try
    Add( AWork ) ;
  finally
    Unlock ;
  end ;
end;

{: Waits for space on the list using the space available semaphore and termination event,
   then if successful returns the next item of work on the list. If the termination
   event gets signalled, returns NIL meaning that the thread should terminate.}
function TWorkList.WaitForWork(const ATerminationEvent: THandle): TWork;
begin
  // wait for work to be available, returning NIL if termination is signalled
  Result := NIL ;
  if not MultiWaitFor( FWorkAvailable, ATerminationEvent ) then exit ;

  // lock the list
  Lock ;
  try
    // safety check - should be unnecessary but it's worthwhile just in case
    if FList.Count > 0 then begin
      // return the first item and remove it from the list
      Result := TWork( FList[ 0 ] ) ;
      FList.Delete( 0 ) ;

      // release the space available semaphore
      SpaceAdded ;
    end ;
  finally
    Unlock ;
  end ;
end;

{: Releases the space available semaphore.}
procedure TWorkList.SpaceAdded;
begin
  ReleaseSemaphore( FSpaceAvailable, 1, NIL ) ;
end;

{: Releases the work available semaphore so that the worker threads
   can wake up and do some work.}
procedure TWorkList.WorkAdded;
begin
  ReleaseSemaphore( FWorkAvailable, 1, NIL ) ;
end;

{: Deletes an item from the work list at the given index.}
procedure TWorkList.Delete(const AIndex: integer);
begin
  Lock ;
  try
    FList.Delete( AIndex ) ;
    SpaceAdded ;
  finally
    Unlock ;
  end ;
end;

end.
